@@ -1,23 +1,29 @@
-# $Id$
+# $Id: Context.pm,v 1.2 2001/06/22 20:42:20 fil Exp $
 package JAAS::Object::Context;
 
 use strict;
 use POE;
 use POE::Component::IKC::Responder;
+use JAAS::Object::Cache::Simple;
 
 sub DEBUG {0};
 
-$JAAS::Object::Context::VERSION="0.1";
+$JAAS::Object::Context::VERSION="0.2";
 
+############################################################
 sub spawn 
 {
-    my($package, $name, $config)=@_;
-    $name||='Context';
+    my($package, $name, $factory, $cache)=@_;
+
+    $name   ||='Context';
+    $factory||=JAAS::Object::Context::SimpleFactory->new();
+    $cache  ||=JAAS::Object::Cache::Simple->new();
+
     POE::Session->create (
-        args=>[$name, $config],
+        args=>[$name, $factory, $cache],
         package_states => [
             $package => [qw(_start register unregister shutdown
-                            create delete
+                            setup clear delete
                             get release putback 
                             add remove
                             list locklist blocklist block do_block unblock)],
@@ -28,13 +34,16 @@ sub spawn
 
 ############################################################
 sub _start {
-    my ($kernel, $heap, $name, $config)=@_[KERNEL, HEAP, ARG0, ARG1];
-    if(ref $config eq 'ARRAY') {
-        my %conf;
-        @conf{@$config}=@$config;
-        $config=\%conf;
+    my($kernel, $heap, $name, $factory, $cache)=
+                    @_[KERNEL, HEAP, ARG0, ARG1, ARG2];
+    if(DEBUG) {
+        die "NO NAME" unless $name;
+        die "NO FACTORY" unless $factory;
+        die "NO CACHE" unless $cache;
     }
-    $heap->{_config}=$config;
+
+    $heap->{_factory}=$factory;
+    $heap->{_cache}=$cache;
 
     DEBUG and print "Context ($name) _start\n";
 
@@ -42,7 +51,7 @@ sub _start {
     POE::Component::IKC::Responder->spawn();
     $kernel->post(IKC=>'publish', 
         $name=> 
-        [qw(create delete 
+        [qw(setup clear delete 
             get release putback 
             add remove
             list locklist blocklist block unblock)]);
@@ -70,34 +79,75 @@ sub unregister {
     # Release all locks held by the quitting client
     my $list = $kernel->call($session, 'list');
     foreach my $id (@$list) {
-        my $locklist = $kernel->call($session, 'locklist', {id=>$id});
-        foreach my $obj (keys %$locklist) {
-            if ($locklist->{$obj} eq $name) {
-                $kernel->yield(release=>{id=>$id, obj=>$obj, name=>$name});
-            } elsif($locklist->{$obj}) {
-                # DEBUG and print "$name isn't $locklist->{$obj}\n";
-            }
-        }
-        if(($heap->{$id}{_blocked_by}||'') eq $name) {
-            $kernel->yield(unblock=>{id=>$id, name=>$name});
-        }
+        # BRUTE FORCE IT!
+        $kernel->yield('clear', {id=>$id, name=>$name});
     }
     DEBUG and print "$name has quit\n";
 }
+
 ###########################################################
-sub create {
-    my ($kernel, $heap, $id)=@_[KERNEL, HEAP, ARG0];
-    $id=$id->{id} if ref $id;
+sub shutdown
+{
+    my($kernel, $session, $heap)=@_[KERNEL, SESSION, HEAP];
+    DEBUG and print "Shutdown of all contexts\n";
+    $session=$session->ID if ref $session;
+    my $list=$kernel->call($session, 'list');
+
+    # remove all aliases
+    foreach my $a ($kernel->alias_list($session)) {
+        $kernel->alias_remove($a);
+    }
+    # retract all published states
+    $kernel->post(IKC=>'retract');
+
+    # equiv to clear() on all our "known" contexts
+    $heap->{_cache}->shutdown($list) if @$list;
+    delete @{$heap}{@$list};
+}
+
+
+
+
+###########################################################
+sub setup
+{
+    my ($kernel, $heap, $args)=@_[KERNEL, HEAP, ARG0];
+    my $id=$args->{id};
+    my $name=$args->{name};
+
     unless (exists $heap->{$id}) {
-        my $ctx=$heap->{$id}={};
-        foreach my $obj (keys %{$heap->{_config}}) {
-            $ctx->{$obj}->{objref} = $heap->{_config}{$obj}->new();
-            $ctx->{$obj}->{lock} = '';  
-        }
-        $ctx->{_block} = 0;
-        DEBUG and print "Context: $id created\n";
+        my $ctx=$heap->{$id}={_block=>0};
+
+        # DEBUG and 
+            print "Context: $id created\n";
     } else {
-        DEBUG and print "Context: $id already exsists\n"
+        DEBUG and print "Context: $id already exists\n"
+    }
+    $heap->{_cache}->setup($id, $name);
+}
+
+###########################################################
+sub clear
+{
+    my ($kernel, $session, $heap, $args)=@_[KERNEL, SESSION, HEAP, ARG0];
+    my $id=$args->{id};
+    my $name=$args->{name};
+    unless (exists $heap->{$id}) {
+        DEBUG and print "Context: $id never existed!\n";
+        return;
+    }
+
+    # REMOVE ALL LOCKS
+    $_[OBJECT]->clear_locks($kernel, $session, $heap, $id, $name);
+
+    # clear all blocking
+    if(($heap->{$id}{_blocked_by}||'') eq $name) {
+        $kernel->yield(unblock=>{id=>$id, name=>$name});
+    }
+
+    # maybe clear from memory?
+    if($heap->{_cache}->clear($id, $name)) {
+        $kernel->yield('delete', $id);
     }
 }
 
@@ -117,37 +167,31 @@ sub delete
             return;
         }
         
+        my $hold=0;
         my $objlist=$kernel->call($session=>'list', $id);
         foreach my $obj (@$objlist) {        # make sure it's not locked
             next unless $ctx->{$obj}{lock};
             DEBUG and print "Context: $id/$obj is locked $ctx->{$obj}{lock}, deffering delete\n";
-            return;
+            $hold=1;
         }
 
-        delete $heap->{$id};
-        DEBUG and print "Context: $id deleted\n";
+        unless($hold) {
+            delete $heap->{$id};
+            # DEBUG and 
+                print "Context: $id deleted\n";
+            $heap->{_cache}->delete($id);
+        }
     } else {
-        DEBUG and print "Context: $id doesn't exsist\n"
+        # DEBUG and 
+        print "Context: $id doesn't exist\n"
     }
 }
 
-###########################################################
-sub shutdown
-{
-    my($kernel, $session, $heap)=@_[KERNEL, SESSION, HEAP];
-    DEBUG and print "Shutdown of all contexts\n";
-    $session=$session->ID if ref $session;
-    foreach my $a ($kernel->alias_list($session)) {
-        $kernel->alias_remove($a);
-    }
-    $kernel->post(IKC=>'retract');
-    $heap={};
-}
-
 
 ###########################################################
 sub add
 {   
+    die "Don't call me!";
     my($heap, $args)=@_[HEAP, ARG0];
     my $id = $args->{id};               # context ID
     my $obj = $args->{obj};             # object name
@@ -167,15 +211,14 @@ sub add
 
         if(($work->{lock}||$name) ne $name) {
             warn "Object $id/$obj isn't locked by $name\n";
-        } else {
-            $work->{objref}=$args->{in};
+            return;
         }
-        return;
+    } else {
+        $work=$heap->{$id}{$obj}||={};        # new object
+        $work->{lock}='';
     }
-
-    $work=$heap->{$id}{$obj}={};        # new object
-    $work->{lock}='';
-    $work->{objref}=$args->{in};
+     
+    $heap->{_cache}->store($obj, $args->{in});
 }
 
 ###########################################################
@@ -198,6 +241,7 @@ sub remove
         warn "Object $id/$obj isn't locked by $name\n";
         return;
     }
+    $heap->{_cache}->delete($id, $obj);
     delete $heap->{$id}{$obj};
 }
 
@@ -205,31 +249,52 @@ sub remove
 sub get 
 {
     my ($kernel, $heap, $args)=@_[KERNEL, HEAP, ARG0];
+    my $id=$args->{id};
+    my $obj=$args->{obj};
+    my $name=$args->{name};
 
-    my $name = $args->{name};
-    DEBUG and print "Object: $args->{id}/$args->{obj} wanted by $name... ";
+    DEBUG and print "Object: $id/$obj wanted by $name... ";
 
-    my $ctx=$heap->{$args->{id}};
-    
-    my $work = $ctx->{$args->{obj}};
-    unless($work->{objref} or $ctx->{_deleted}) { # ERR -1: Obj. doesn't exist
-        DEBUG and print "doesn't exists\n";
+    my $ctx=$heap->{$id};
+    unless($ctx) {
+        DEBUG and print "context doesn't exist\n";
+        return -1;
+    }
+    if($ctx->{_deleted}) {
+        DEBUG and print "context deleted";
         return -1;
     }
 
-    unless(($work->{lock} || $name) eq $name) {
-        DEBUG and print "was locked by $work->{lock}!\n";
+    unless(($ctx->{$obj}{lock} || $name) eq $name) {
+        DEBUG and print "was locked by $ctx->{$obj}{lock}!\n";
         return 0;                                   # ERR 0: Obj. locked 
     }
-
     unless(($ctx->{_blocked_by} || $name) eq $name) {    # ERR -2: blocked
         DEBUG and print "context blocked by $ctx->{_blocked_by}!\n";
         return -2;
     }
 
-    $work->{lock} ||= $name;
+    ######################################################
+    my $out=$heap->{_cache}->fetch($id, $obj);
+    unless($out) {
+        DEBUG and print "creating... ";
+        $out=$heap->{_factory}->make($obj);
+        unless($out) {
+            DEBUG and print "can't!\n";
+            return -1;
+        }
+        $heap->{_cache}->store($id, $obj, $out) 
+    }
+
+    unless($out) {                              # ERR -1: Obj. doesn't exist
+        DEBUG and print "doesn't exists\n";
+        return -1;
+    }
     DEBUG and print "OK\n";
-    return $work->{objref};
+
+    $_[OBJECT]->get_lock($kernel, $heap, $id, $obj, $name);
+    
+    return $out;
 }
 
 ###########################################################
@@ -246,20 +311,15 @@ sub release
         return;
     }
     unless($work) {
-        DEBUG and print "Object $id/$args->{obj} doesn't exist\n";
+        DEBUG and print "Object: $id/$args->{obj} doesn't exist\n";
         return;
     }
     if ($work->{lock} ne $name) {
-        DEBUG and print "Object $id/$args->{obj} isn't locked by $name\n";
+        DEBUG and print "Object: $id/$args->{obj} isn't locked by $name\n";
         return;
     }
 
-    $work->{lock}='';
-    $work->{objref}=$args->{in} if $args->{in};
-    DEBUG and print "Object: $id/$args->{obj} released for $name\n";
-
-    $kernel->yield('do_block', $id) if $ctx->{_blocked_by};
-    $kernel->yield('delete', $id)   if $ctx->{_deleted};
+    $_[OBJECT]->release_lock($kernel, $heap, $id, $args->{obj}, $name);
 }
 
 ###########################################################
@@ -268,24 +328,67 @@ sub putback {
     my $name = $args->{name};                   # remote kernel
     my $id = $args->{id};                       # context ID
     my $ctx = $heap->{$id};                     # Context
-    my $work = $ctx->{$args->{obj}};            # object
+    my $obj = $args->{obj};                     # object name
+    my $work = $ctx->{$obj};                    # object
 
     unless($ctx) {
         DEBUG and print "Context $id doesn't exist\n";
         return;
     }
-    unless($work) {
-        DEBUG and print "Object $id/$args->{obj} doesn't exist\n";
+    unless(exists $ctx->{$obj}) {
+        DEBUG and print "Object: $id/$args->{obj} doesn't exist\n";
         return;
     }
     unless($work->{lock} eq $name) {
-        DEBUG and print "Object $id/$args->{obj} isn't locked by $name\n";
+        DEBUG and print "Object: $id/$args->{obj} isn't locked by $name\n";
         return;
     }
 
     DEBUG and print "Object: $id/$args->{obj} changed by $name\n";
-    $work->{objref} = $args->{in};
-    $kernel->yield('release', {id=>$id, obj=>$args->{obj}, name=>$name});
+    
+    $heap->{_cache}->store($id, $obj, $args->{in});
+
+    $_[OBJECT]->release_lock($kernel, $heap, $id, $obj,  $name);
+}
+
+
+
+
+###########################################################
+sub get_lock
+{
+    my($package, $kernel, $heap, $id, $obj, $name)=@_;
+    DEBUG and print "Lock: $id/$obj reserved for $name\n";
+
+    $heap->{$id}{$obj}{lock} ||= $name;
+}
+
+###########################################################
+sub release_lock
+{
+    my($package, $kernel, $heap, $id, $obj, $name)=@_;
+    my $ctx=$heap->{$id};
+    $ctx->{$obj}{lock}='';
+
+    DEBUG and print "Lock: $id/$obj released for $name\n";
+
+    $kernel->yield('do_block', $id) if $ctx->{_blocked_by};
+    $kernel->yield('delete', $id)   if $ctx->{_deleted};
+}
+
+###########################################################
+sub clear_locks
+{
+    my($package, $kernel, $session, $heap, $id, $name)=@_;
+    # clear all locks
+    my $locklist = $kernel->call($session, 'locklist', {id=>$id});
+    foreach my $obj (keys %$locklist) {
+        if ($locklist->{$obj} eq $name) {
+            $package->release_lock($kernel, $heap, $id, $obj, $name);
+        } elsif($locklist->{$obj}) {
+            # DEBUG and print "$name isn't $locklist->{$obj}\n";
+        }
+    }
 }
 
 
@@ -293,9 +396,7 @@ sub putback {
 sub list {
     my ($kernel, $heap, $id)=@_[KERNEL, HEAP, ARG0];
     
-    return [grep !/^_/, keys %$heap] unless $id;
-
-    return [grep !/^_/, keys %{$heap->{$id}}];
+    return [grep !/^_/, keys %{$id ? $heap->{$id} : $heap}];
 }
     
 ###########################################################
@@ -319,6 +420,8 @@ sub locklist {
     return \%locklist;
 }
 
+
+
 ###########################################################
 sub blocklist {
     my ($kernel, $session, $heap, $args)=@_[KERNEL, SESSION, HEAP, ARG0];
@@ -419,7 +522,33 @@ sub unblock {
         $kernel->yield('delete', $id) if $ctx->{_deleted};
     }
 }
-            
+
+
+
+
+##############################################################################
+package JAAS::Object::Context::SimpleFactory;
+use strict;
+
+sub new { bless {}, $_[0]}
+use Carp;
+sub make
+{
+    my($self, $name)=@_;
+    return if exists $self->{deleted}{$name};
+    return eval{$name->new()};
+}
+sub delete
+{
+    my($self, $name)=@_;
+    $self->{deleted}{$name}=1;
+}
+
+
+
+
+
+
 1;
 __END__
 
@@ -430,7 +559,7 @@ JAAS::Object::Context - JAAS object repository
 =head1 SYNOPSIS
 
     # on the server side:
-    JAAS::Object::Context->spawn('Context', {Cart=>"My::Shop::Cart"});
+    JAAS::Object::Context->spawn('Context', $factory);
 
     # On the client side, subscribe to the Context session in the server
     # aka poe://Server/Context;
@@ -477,18 +606,24 @@ JAAS::Object::Context - JAAS object repository
 =head1 DESCRIPTION
 
 JAAF::Object::Context is a POE IKC server acting as an object repository.
-Clients connect to it, create working contexts, get objects, work on them
-and put them back in the repository. While a client has an object, a lock is
-registered under its name, preventing anyone else from getting the object.
-When a client disconnects, all its locks are automatically released.
+Clients connect to it, setup working contexts, get objects, work on them and
+put them back in the repository, then clear the context.  While a client has
+an object, a lock is registered under its name, preventing anyone else from
+getting the object.  When a client disconnects, all its locks are
+automatically released and all interest is cleared.
+
+Being "interested" is a silly pattern : when an external session wants to
+use a context, it first posts 'setup'.  The server will create the context
+if needed.  The session will then play with objects in the session as
+needed.  When done, the session calles 'clear' to revoke it's interest. 
+When no one is interested in the context, it is deleted from memory.
 
 The 'working context' is a collection of objects, always the same, as
-defined in a config file (currently not implemented). For example, an e-comm config
-could define objects 'Cart', 'User' and 'Catalog'; any new context would
-create new instances of these objects under a particular context name.
-Parameters for their new() constructor, if any, should be specified in the
-config file. The client needs a 'use' statement for every object imported
-from Context, so it can use its methods.
+created by C<$factory>.  Each context is maintained by the C<$cache>. For
+example, an e-comm config could define objects 'Cart', 'User' and 'Catalog';
+any new context would create new instances of these objects as needed (via
+C<$factory>) and store in C<$cache> under a particular context name. 
+Persistance is implemented via the C<$cache>.
 
 Because no magic is done about syncing objects across proccess,
 JAAS::Object::Context should only be spawned in a non-forking server.
@@ -498,33 +633,52 @@ JAAS::Object::Context should only be spawned in a non-forking server.
 
 =head2 spawn
 
-    JAAS::Object::Context->spawn($name, $config);
+    JAAS::Object::Context->spawn($name, $factory, $cache);
 
 Creates the a session to act at the object repository.  C<$name> is the name
-of the session.  C<$config> is the object that will be used to create and
-configure objects in the contexts.  
+of the session.  C<$factory> is an object that will be used to create and
+configure objects in the contexts.  If it is omited, a very simple object is
+used, that assumes that object names are packages and calls C<new> on them.
+C<$cache> is an object that will be used to store objects related to a
+context.  If it is omited, a very simple object is used, that keeps
+everything in a hash and deletes all objects when no one is interested in the
+context anymore.
 
 
 =head1 STATES
 
-=head2 create
+=head2 setup
 
-    $kernel->call('IKC', 'post', 'poe://Server/Context/create', $ID);
+    $kernel->call('IKC', 'post', 'poe://Server/Context/setup', 
+                {id=>$ID, name=>$me});
+
+C<$me> wants to use a context called C<$ID>.  A new context is created if it 
+doesn't exist.  Context is kept around until everyone has "cleared" their
+interest.  This is a crude form of GC.
+
+=head2 clear
+
+    $kernel->call('IKC', 'post', 'poe://Server/Context/setup', 
+                {id=>$ID, name=>$me});
+
+C<$me> no longer wants to use a context called C<$ID>.  This means that the
+repository can get it out of memory and (depending on the cache).
 
-Creates a new working context called C<$ID> on Context. This creates
-new holder of objects as specified in the initial config file which Context
-loads at startup. If the context ID already exists, C<create> does nothing.
 
 =head2 delete
 
     $kernel->post('IKC', 'post', 'poe://Server/Context/delete', $ID)
 
+Purposefully delete a context.  This holds until all blocks and locks are
+removed, including your own.
+
 
 =head2 shutdown
 
     $kernel->post(Context=>'shutdown');
 
-Removes the session for the kernel.
+Removes the Context session from the kernel.  This also calls C<shutdown()>
+on the cache.
 
 =head2 get
 
@@ -534,7 +688,7 @@ Removes the session for the kernel.
 Gets object named C<$obj> from context C<$ID> on Context, then passed on to
 the callback handler. C<get> locks the object under the name C<$name>. If
 the object is already locked or if the entire context was blocked, C<get>
-returns 0 to callback.  If the object doesn't exsist, or the context, even,
+returns 0 to callback.  If the object doesn't exist, or the context doesn't,
 callback is called with -1.
 
 =head2 putback
@@ -542,32 +696,34 @@ callback is called with -1.
         $kernel->post('IKC', 'post', 'poe://Server/Context/putback',
                 {id=>$ID, obj=>$obj, name=>$name, in=>$frozen_object});
 
-Gives the object back to Context. Object must be previously frozen using
-POE::Filter::Reference. A C<putback> automatically unlocks the object.
+Gives the object back to Context.  A C<putback> automatically unlocks the
+object.
 
 =head2 release
 
         $kernel->post('IKC', 'post', 'poe://Server/Context/release',
                 {id=>$ID, obj=>$obj, name=>$name});
 
-Releases the lock on an object. This handler is preferred to C<putback> if
+Releases the lock on an object. This message is preferred to C<putback> if
 the client hasn't modified anything on the object.
 
 =head2 list
 
         $kernel->post('IKC', 'post', 'poe://Server/Context/list',
-                '', 'callback');
+                [$id], 'callback');
 
 Fetches a list of active context IDs on Context, as an arrayref, to the
-callback handler.
+callback handler.  If C<$id> is given, fetches a list of active objects in
+that context.
 
 =head2 locklist
 
     $kernel->call('IKC', 'call', 'poe://Server/Context/locklist',
          {id=>$ID, type=>$type}, 'callback');
 
-Gives a list of objects contained in C<$ID>, as a hashref, to the callback
-handler. There are two types of lists, as specified by the C<type> argument.
+Gives a list of active objects contained in C<$ID>, as a hashref, to the
+callback handler. There are two types of lists, as specified by the C<type>
+argument.
 
 =item Type 0
 
@@ -614,20 +770,16 @@ after an 'unblock'.
 
 These are the guidelines for the clients connecting to the Context server.
 
-- Clients must C<use> the packages for the objects requested from Context,
-so they can use their methods.
+- If the client is in a different process, it must C<use> the packages for
+the objects requested from Context, so it can use their methods.
 
-- A client should create for itself a unique name, like 'C<catalog-lookup$$>' for
-example. That name _must_ be passed in the subscription phase, in the
-C<create_ikc_client> method, as it will be used for unlocking objects when
-the client disconnects.
+- A client should create for itself a unique name, like
+'C<catalog-lookup$$>' for example. That name _must_ be passed in the
+subscription phase, in the C<create_ikc_client> method, as it will be used
+for unlocking objects when the client disconnects.
 
-- Don't forget to C<use POE::Filter::Reference>, as C<get> and C<putback> will
-need it.
-
-- Never use names starting with underscore (_) for ID and client
-names, these are reserved names for Context. Unless you're a hacker. A good
-one.
+- Never use names starting with underscore (_) for ID and client names,
+these are reserved names for Context. Unless you're a hacker. A good one.
 
 =head1 BUGS
 
@@ -644,4 +796,16 @@ Philip Gwyn <gwynp@artware.qc.ca>
 
 =cut
 
-$Log$
+$Log: Context.pm,v $
+Revision 1.2  2001/06/22 20:42:20  fil
+Added Factory, Cache/*
+Reworked internals... Still have problems, though.
+
+Revision 1.1  2001/06/21 04:11:19  fil
+Totaly rearranged things:
+    Context.pl is now Context.pm and server.perl
+    Blocking and locking now are revoked gracefully when remote closes
+    Blocking loop is now unrolled.
+    Removed all the delay()s.  When something is deffered, we check again
+    when the condition is encountered.
+
